﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Timers;
using System.Threading;

namespace STC89C5xProgrammer
{
    public partial class SerialPortForm : Form
    {
        public volatile int mode = 0;

        private Form1 parent = null;
        private System.Timers.Timer timer = null;
        private volatile int tries = 0;
        
        private string txString = "";
        private byte[] rxBuffer = null;
        private int rxInIndex = 0;
        private string rxString = "";

        private bool displayInfoOnly = true;
        private bool setHighSpeedMode = false;

        public SerialPortForm(Form1 parentForm)
        {
            InitializeComponent();
            parent = parentForm;
            Icon = parent.Icon;

            timer = new System.Timers.Timer();
            timer.Elapsed += timerEvent;
            timer.Enabled = false;
            timer.AutoReset = true;

            rxBuffer = new byte[32768];
            rxInIndex = 0;
            mode = 0;
        }

        private void transmit(byte[] buffer, int offset, int count)
        {
            for (int i = 0; i < count; i++)
            {
                txString += buffer[i].ToString("X2") + " ";
            }
            Utils.SetControlPropertyValue(txLabel, "Text", txString);
            serialPort1.Write(buffer, offset, count);
        }

        private void transmitPacket(byte messageType, int length, byte[] body)
        {
            byte[] buffer = new byte[length + 8];

            buffer[0] = 0x46;
            buffer[1] = 0xB9; // package header 
            buffer[2] = 0x6A; // package send flag
            buffer[3] = 0x00;
            buffer[4] = (byte)(length + 6);
            buffer[5] = messageType;

            for (int i = 0; i < length; i++)
            {
                buffer[6 + i] = body[i];
            }

            byte checksum = 0x01;
            for (int i = 0; i < length + 6; i++)
            {
                checksum += buffer[i];
            }
            buffer[6 + length] = checksum;

            buffer[7 + length] = 0x16; // end of message marker

            txString += "\n"; // new line to make packets more visible
            rxString += "\n";

            rxInIndex = 0;
            transmit(buffer, 0, length + 8);
        }

        private void checkTargetInfo() // check initial message received from target chip
        {
            int head = -1;
            for (int i = 0; i < rxInIndex - 3; i ++)
            {
                // if (rxBuffer[i] == 0x68 && rxBuffer[i+1] == 0x00 && (rxBuffer[i+2] == 0x3A || rxBuffer[i + 2] == 0x3B) && rxBuffer[i+3] == 0x00)
                if (rxBuffer[i] == 0x68 && rxBuffer[i + 1] == 0x00  && rxBuffer[i + 3] == 0x00)
                {
                    // dev board chip returns 68 00 3A 00.  POV chip returns 68 00 3B 00 ?
                    head = i;
                    break;
                }
            }
            if (head < 0) return;
            
            // wait for a terminated (0x16) packet that's long enough to be a target info message
            if (rxBuffer[rxInIndex - 1] != 0x16 || head + 25 >= rxInIndex)
            {
                return;
            }

            // next 8 pairs of bytes are autobaud measurement - find average, perform whacky calculation to work out crystal frequency
            Int32 k = 0;
            for (int i = head + 4; i < head + 20; i += 2)
            {
                k += 256 * rxBuffer[i] + rxBuffer[i + 1];
            }
            bool highSpeed = (rxBuffer[head + 22] & 0x01) == 0;
            double denominator = highSpeed ? 9.305600 : 4.625800;
            int crystalHz = (int)((double)k * (double)serialPort1.BaudRate / denominator);
            parent.ciForm.crystalMHz = crystalHz / 1000000.0;
            parent.ciForm.bslVersion = rxBuffer[head + 20] * 256 + rxBuffer[head + 21];
            parent.ciForm.optionsByte = rxBuffer[head + 22];
            parent.ciForm.magic = rxBuffer[head + 23] * 256 + rxBuffer[head + 24];
            parent.ciForm.update();
            
            byte checksum = 0x00;
            for (int i = head; i < rxInIndex - 2; i++)
            {
                checksum += rxBuffer[i];
            }
            if (checksum != rxBuffer[rxInIndex - 2])
            {
                rxString += "\nProblem - bad checksum on Target info";
                Utils.SetControlPropertyValue(rxLabel, "Text", rxString);
                mode = 0;
                return;
            }

            if (displayInfoOnly)
            {
                mode = int.MaxValue;
                return;
            }

            // after receiving target info, send a change baudrate message (code 0x8F)  for now assume operation at 9600 baud with an 11.0592 crystal
            mode = 3;
            // byte[] checkBaud9600_11M = { 0xFF, 0xDC, 0x00, 0x48, 0x28, 0x81 };
            // transmitPacket(0x8F, 6, checkBaud9600_11M);
            byte[] changeBaud = createBaudData(highSpeed, crystalHz, serialPort1.BaudRate);
            transmitPacket(0x8F, 6, changeBaud);
        }

        private int highestDataPacketAddress; // send the highest packet first, then packets from beginning until all have been sent once
        private int nextDataPacketAddress;

        private void handleReceivedPacket(byte messageType, int length, byte[] body)
        {
            
            rxString += "\nmessageType " + messageType.ToString("X2") + " length " + length.ToString();
            Utils.SetControlPropertyValue(rxLabel, "Text", rxString);
            switch (messageType)
            {
                case 0x80: // received ACK response
                    if (mode < 9) // send 5 ACKS
                    {
                        mode++;
                        byte[] ANOTHER_ACK = { };
                        transmitPacket(0x80, 0, ANOTHER_ACK);
                        break;
                    }
                    else if (mode == 9) // send a RDY
                    {
                        mode = 10;
                        highestDataPacketAddress = parent.memory.Length - 128;
                        nextDataPacketAddress = highestDataPacketAddress;
                        byte[] RDY = { 0x02 };
                        transmitPacket(0x84, 1, RDY);
                    }
                    else // sending memory contents
                    {
                        ++mode;
              
                        if (nextDataPacketAddress > highestDataPacketAddress) // finished sending memory
                        {
                            byte[] SET_OPTIONS = { 0, 0xFF, 0xFF, 0xFF };
                            SET_OPTIONS[0] = (byte)((parent.ciForm.optionsByte & 0xFE) | (setHighSpeedMode ? 0x00 : 0x01)); // LSB set gives low (12T) speed
                            transmitPacket(0x8D, 4, SET_OPTIONS);
                        }
                        else // send a 128-byte packet
                        {
                            byte[] buffer = new byte[134];
                            buffer[0] = 0x00;
                            buffer[1] = 0x00;
                            buffer[2] = (byte)(nextDataPacketAddress >> 8);
                            buffer[3] = (byte)(nextDataPacketAddress & 0x00FF);
                            buffer[4] = 0x00;
                            buffer[5] = 0x80;
                            for (int i = 0; i < 128; i++)
                            {
                                buffer[6 + i] = parent.memory[nextDataPacketAddress + i];
                            }
                            transmitPacket(0x00, 134, buffer);
                        }

                        if (nextDataPacketAddress == highestDataPacketAddress) // initial packet sent
                        {
                            if (highestDataPacketAddress == 0) // if there was only one packet (unlikely)
                            {
                                nextDataPacketAddress += 128; // force next address too high so as to finish
                            }
                            else
                            {
                                nextDataPacketAddress = 0; // now start from the bottom of memory
                            }
                        }
                        else // keep sending next packet till we're done
                        {
                            nextDataPacketAddress += 128;
                            if (nextDataPacketAddress == highestDataPacketAddress)
                            {
                                nextDataPacketAddress += 128; // force next address too high so as to finish
                            }
                        }
                    }
                    break;
                case 0x8D: // received response to set options message (code 0x8D) - send end message
                    byte[] END_MSG = { };
                    transmitPacket(0x82, 0, END_MSG);
                    return;
                case 0x8E: // received response to check baudrate message (code 0x8E) - next send first of five ACK messages (don't know why yet)
                    mode = 5;
                    byte[] ACK = { };
                    transmitPacket(0x80, 0, ACK);
                    break;
                case 0x8F: // received response to change baudrate message (code 0x8F)
                    mode = 4;
                    // byte[] changeBaud960011M = { 0xFF, 0xDC, 0x00, 0x48, 0x17 };
                    transmitPacket(0x8E, 5, body); // checkbaud is same as response to changebaud with last byte removed
                    break;
            }
        }

        private void checkIfPacketReceived()
        {
            // wait for a terminated (0x16) packet where the length buffer[4] matches the received length
            if (rxInIndex < 8 || rxBuffer[rxInIndex - 1] != 0x16 || rxBuffer[4] != rxInIndex - 2)
            {
                return;
            }
            // check that the packet has a header followed by a receive flag
            if (rxBuffer[0] != 0x46 || rxBuffer[1] != 0xB9 || rxBuffer[2] != 0x68)
            {
                rxString += "\nProblem - received packet did not begin with 46 B9 68";
                Utils.SetControlPropertyValue(rxLabel, "Text", rxString);
                mode = 0;
                return;
            }

            byte checksum = 0x01;
            for (int i = 0; i < rxInIndex - 2; i++)
            {
                checksum += rxBuffer[i];
            }

            if (checksum != rxBuffer[rxInIndex - 2])
            {
                rxString += "\nProblem - bad checksum on received packet";
                Utils.SetControlPropertyValue(rxLabel, "Text", rxString);
                mode = 0;
                return;
            }

            byte[] body = new byte[rxBuffer[4] - 6];
            for (int i = 0; i < rxBuffer[4] - 6; i++)
            {
                body[i] = rxBuffer[6 + i];
            }

            handleReceivedPacket(rxBuffer[5], rxBuffer[4] - 6, body);
        }

        private void serialPort1_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            int bytesToRead = serialPort1.BytesToRead;
            serialPort1.Read(rxBuffer, rxInIndex, bytesToRead);
            for (int i = 0; i < bytesToRead; i++)
            {
                rxString += rxBuffer[rxInIndex++].ToString("X2") + " ";
            }
            Utils.SetControlPropertyValue(rxLabel, "Text", rxString);
            if (mode == 1) // receiving response to initial '7F' - should be information about the chip
            {
                checkTargetInfo();
            }
            else
            {
                checkIfPacketReceived();
            }
        }


        private void timerEvent(object sender, ElapsedEventArgs e)
        {
            byte[] buffer = new byte[1];

            switch (mode)
            {
                case 1: // attempting to connect
                    if (++tries > 400)
                    {
                        mode = 0;
                    }
                    
                    buffer[0] = 0x7F;
                    transmit(buffer, 0, 1);
                    break;
                default:
                    timer.Stop();
                    break;
            }
        }

        private byte[] createBaudData(bool highSpeed, int Hz, int baud)
        {
            byte[] baudData = new byte[6];

            double sampleRate = highSpeed ? 16.0 : 32.0;
            double brt = 65536.0 - (((double)Hz) / (baud * sampleRate));
            int brtInt = (int)(brt + 0.5);
            baudData[0] = (byte)(brtInt / 256);
            baudData[1] = (byte)(brtInt & 255);
            baudData[2] = (byte)(255 - baudData[0]);
            baudData[3] = (byte)((2 * (256 - baudData[1])) & 255);
            baudData[4] = 0xA0;
            byte iap = 0x80;
            if (Hz < 5000000) iap = 0x83;
            else if (Hz < 10000000) iap = 0x82;
            else if (Hz < 20000000) iap = 0x81;
            baudData[5] = iap;
            return baudData;
        }

        public void go(string portName, int baud, bool uploadProgram, bool setHighSpeed)
        {
            if (mode != 0) return;

            displayInfoOnly = !uploadProgram;
            setHighSpeedMode = setHighSpeed;

            txLabel.Text = "";
            rxLabel.Text = "";

            if (serialPort1.IsOpen)
            {
                serialPort1.Close();
            }

            serialPort1.PortName = portName;
            serialPort1.BaudRate = baud; // chip auto detects initial baud.  works well at 9600
            serialPort1.DataBits = 8;
            serialPort1.Parity = System.IO.Ports.Parity.None;
            serialPort1.StopBits = System.IO.Ports.StopBits.One;
            serialPort1.Open();

            txString = "";
            rxString = "";
            rxInIndex = 0;
            txLabel.Text = "";
            rxLabel.Text = "";

            mode = 1; // attempt to connect
            tries = 0; // count of attempts
            timer.Interval = 20; // try fifty times per second
            timer.Enabled = true;
        }

        private void SerialPortForm_Resize(object sender, EventArgs e)
        {
            if (Width > 80)
            {
                txLabel.Width = Width - 40;
                rxLabel.Width = Width - 40;
            }
            else
            {
                txLabel.Width = 40;
                rxLabel.Width = 40;
            }

            if (Height > 137)
            {
                txLabel.Height = (Height - 57) / 2;
                rxLabel.Height = (Height - 57) / 2;
            }
            else
            {
                txLabel.Height = 40;
                rxLabel.Height = 40;
            }
            rxLabel.Top = txLabel.Top + txLabel.Height;
        }

        private void SerialPortForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            timer.Stop();
            timer.Dispose();

            if (serialPort1.IsOpen)
            {
                e.Cancel = true; //cancel the form closing
                Thread CloseDown = new Thread(new ThreadStart(CloseSerialOnExit));
                CloseDown.Start(); //close port in new thread to avoid hang
            }
        }

        private void CloseSerialOnExit()
        {
            try
            {
                serialPort1.Close();
            }

            catch (Exception ex)
            {
                MessageBox.Show(ex.Message); //catch any serial port closing error messages
            }
            try
            {
                this.Invoke(new EventHandler(NowClose)); //now close back in the main thread
            }
            catch (InvalidOperationException)
            {
            }
        }

        private void NowClose(object sender, EventArgs e)
        {
            this.Close(); //now close the form
        }
    }
}
